### SCRIPT EXPLORACIÓN
# Autores: Gabriel Abilleira, Paula Álvarez y Javier Godos 
# Equipo: E-No_Logos (UC3M)

### IMPORTANDO LOS PAQUETES
import matplotlib.pyplot as plt
plt.style.use("ggplot")
import pandas as pd
import numpy as np
import missingno as msno
from scipy import stats
import math
from sklearn.experimental import enable_iterative_imputer
from sklearn.impute import KNNImputer, SimpleImputer, IterativeImputer
from sklearn.pipeline import Pipeline, make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split
from sklearn.model_selection import cross_val_score, KFold, RepeatedKFold
from sklearn.metrics import mean_squared_error
from sklearn.feature_selection import SelectKBest, f_regression, mutual_info_regression
import optuna
import numpy as np
from lazypredict.Supervised import LazyRegressor
from lightgbm import LGBMRegressor
import xgboost as xgb
import seaborn as sns
from sklearn.preprocessing import PolynomialFeatures
import pathlib
from sklearn.ensemble import ExtraTreesRegressor, RandomForestRegressor, GradientBoostingRegressor, StackingRegressor, HistGradientBoostingRegressor
from sklearn.linear_model import LinearRegression, PassiveAggressiveRegressor
import random
from sklearn import compose
from sklearn.neighbors import KNeighborsRegressor
import dalex as dx

# Definición de la semilla para reproducibilidad
seed_num = 669280

############################# OBTENCIÓN DF #############################

# Cargamos los datos

# Los datos se encuentran en la carpeta actual, si no, hay que poner ahí el path de la carpeta con los datos
current_folder = str(pathlib.Path().resolve())

# Datos de las viñas:
UH_2023_TRAIN = pd.read_csv((current_folder+"\\UH_2023_TRAIN.txt")
                            , sep = "|", encoding = "utf-8")

# Datos de la información meteorológica sin agregar.    
DATOS_METEO = pd.read_csv((current_folder+"\\DATOS_METEO.txt")
                            , sep = "|", encoding = "utf-8", parse_dates = [0])

# Datos meteorológicos agregados:
DATOS_ETO = pd.read_csv((current_folder+"\\DATOS_ETO.txt")
                            , sep = "|", encoding = "utf-8")

### DEFINICIÓN DE FUNCIONES
def getPhase(x):
    
    """
    Esta función redibe una fila (x) de un dataframe de pandas y devuelve la
    fase del ciclo de la viña a la que pertenece, en función de la fecha en la
    que se recogieron los datos. Finalmente, si la fecha es superior al 30 de junio de 
    ese año, señala que esos datos deberían utilzarse en la producción del año siguiente.
    """
    
    day = x["Day"]
    month = x["Month"]
    year = x["Year"]
    if day < 10:
        day = "0" + str(day).split(".")[0]
    else:
        day = str(day).split(".")[0]
        
    month = str(month).split(".")[0]
    
    key = int(month + day)
    
    if key >= 305 and key <= 323:
        return "L", int(year)
    elif key >= 324 and key <= 520:
        return "B", int(year)
    elif key >= 521 and key <= 630:
        return "F1", int(year)
    elif key >= 701 and key <= 810:
        year+= 1
        return "F2", int(year)
    elif key >= 811 and key <= 931:
        year += 1
        return "M", int(year)
    elif key >= 1001 and key <= 1115:
        year += 1
        return "C", int(year)
    else:
        if key >= 1116 and key <= 1231:
            year += 1
        return "R", int(year)
    
    
def getHeight(x):
    
    """
    Esta función recibe una fila (x) de un dataframe de pandas, y devuelve los datos de la columna ALTITUD en el
    formato adecuado para ajustar el modelo y realizar las predicciones.
    """
    
    if type(x["ALTITUD"]) == str:
        alt = [float(i) for i in x["ALTITUD"].split("-")]
        if len(alt) > 2:
            return (alt[0] + alt[1])/2
        else:
            return alt[0]
    else:
        return x["ALTITUD"]

### CARGANDO Y TRANSFORMANDO LOS DATASETS

## EDA 
print(UH_2023_TRAIN.info())

print(UH_2023_TRAIN.sort_values(["ID_FINCA","CAMPAÑA"]).head(20))
print(UH_2023_TRAIN.CAMPAÑA.unique())

print(UH_2023_TRAIN[[i for i in UH_2023_TRAIN if i not in ["CAMPAÑA", "ID_FINCA", "ID_ZONA", "ID_ESTACION"]]].describe())

## Transformamos los datos
UH_2023_TRAIN.SUPERFICIE.replace(0.0, np.nan, inplace = True)
UH_2023_TRAIN.ALTITUD = list(map(getHeight, [row for i,row in UH_2023_TRAIN.iterrows()]))

# Información sobre la superficie de las fincas en base a diferentes variables de uva:
INFO_SUPERFICIE = UH_2023_TRAIN[["ID_FINCA", "VARIEDAD", "MODO", "TIPO", "SUPERFICIE"]].sort_values(["ID_FINCA","VARIEDAD", "MODO", "TIPO"]).dropna().drop_duplicates().reset_index().drop("index", axis = 1)

# Imputando los valores de la superficie, en base a diferentes variables de la uva, si están disponibles:

todo = 0 #Número de fincas imputadas que mantuvieron Variedad, Modo y Tipo de la uva.
part1 = 0 #Número de fincas imputadas que mantuvieron Variedad y Modo de la uva.
part2 = 0 #Número de fincas imputadas que mantuvieron Variedad de la uva.
part3 = 0 #Número de fincas imputadas que no mantuvieron ninguna variable de la uva.

for i,row in UH_2023_TRAIN.iterrows():
    if math.isnan(row["SUPERFICIE"]):
        sup_value = list(INFO_SUPERFICIE.loc[(INFO_SUPERFICIE.ID_FINCA == row["ID_FINCA"]) & (INFO_SUPERFICIE.VARIEDAD == row["VARIEDAD"]) & (INFO_SUPERFICIE.MODO == row["MODO"]) & (INFO_SUPERFICIE.TIPO == row["TIPO"]), "SUPERFICIE"])
        if len(sup_value) > 0:
            todo+=1
            if len(sup_value) > 1:
                UH_2023_TRAIN.loc[i, "SUPERFICIE"] = np.mean(sup_value) #Si hay varias coincidencias, imputamos la media de la superficie
            else:
                UH_2023_TRAIN.loc[i, "SUPERFICIE"] = sup_value[0]
        else:
            sup_value = list(INFO_SUPERFICIE.loc[(INFO_SUPERFICIE.ID_FINCA == row["ID_FINCA"]) & (INFO_SUPERFICIE.VARIEDAD == row["VARIEDAD"]) & (INFO_SUPERFICIE.MODO == row["MODO"]), "SUPERFICIE"])
            if len(sup_value) > 0:
                part1+=1
                if len(sup_value) > 1:
                    UH_2023_TRAIN.loc[i, "SUPERFICIE"] = np.mean(sup_value)
                else:
                    UH_2023_TRAIN.loc[i, "SUPERFICIE"] = sup_value[0]
            else:
                sup_value = list(INFO_SUPERFICIE.loc[(INFO_SUPERFICIE.ID_FINCA == row["ID_FINCA"]) & (INFO_SUPERFICIE.VARIEDAD == row["VARIEDAD"]), "SUPERFICIE"])
                if len(sup_value) > 0:
                    part2+=1
                    if len(sup_value) > 1:
                        UH_2023_TRAIN.loc[i, "SUPERFICIE"] = np.mean(sup_value)
                    else:
                        UH_2023_TRAIN.loc[i, "SUPERFICIE"] = sup_value[0]
                else:
                    sup_value = list(INFO_SUPERFICIE.loc[(INFO_SUPERFICIE.ID_FINCA == row["ID_FINCA"]), "SUPERFICIE"])
                    if len(sup_value) > 0:
                        part3+=1
                        if len(sup_value) > 1:
                            UH_2023_TRAIN.loc[i, "SUPERFICIE"] = np.mean(sup_value)
                        else:
                            UH_2023_TRAIN.loc[i, "SUPERFICIE"] = sup_value[0]
                    else:
                        UH_2023_TRAIN.loc[i, "SUPERFICIE"] = np.nan #Si no tenemos datos para la superficie de la finca, imputamos un NaN.
    else:
        pass
    

# Imputando los valores de la altitud, en base al ID de la estación meteorológica de la finca:
    
altitud = UH_2023_TRAIN.groupby("ID_ESTACION").agg({"ALTITUD":np.nanmedian}).reset_index()
for i, row in UH_2023_TRAIN.iterrows():
    if math.isnan(row["ALTITUD"]):
        alt_val = altitud.loc[altitud.ID_ESTACION == row["ID_ESTACION"], "ALTITUD"].values
        if len(alt_val) > 0:
            UH_2023_TRAIN.loc[i, "ALTITUD"] = alt_val[0]
        else:
            pass
    else:
        pass
    
# Creamos las variables de los retardos de la producción, para 1 y 2 años menos, respectivamente
lag_prod_values = []
for i, row in UH_2023_TRAIN.iterrows():
    campaña = row["CAMPAÑA"] - 1
    finca = row["ID_FINCA"]
    variedad = row["VARIEDAD"]
    lag_prod = UH_2023_TRAIN.loc[(UH_2023_TRAIN.CAMPAÑA == campaña) & (UH_2023_TRAIN.ID_FINCA == finca) & (UH_2023_TRAIN.VARIEDAD == variedad)
                                 , "PRODUCCION"].values
    if len(lag_prod) > 0:
        if len(lag_prod) == 1:
            lag_prod_values.append(lag_prod[0])
        else:
            lag_prod_values.append(np.mean(lag_prod))
    else:
        lag_prod_values.append(np.nan)
        
lag_prod_values_2 = []
for i, row in UH_2023_TRAIN.iterrows():
    campaña = row["CAMPAÑA"] - 2
    finca = row["ID_FINCA"]
    variedad = row["VARIEDAD"]
    lag_prod = UH_2023_TRAIN.loc[(UH_2023_TRAIN.CAMPAÑA == campaña) & (UH_2023_TRAIN.ID_FINCA == finca) & (UH_2023_TRAIN.VARIEDAD == variedad)
                                 , "PRODUCCION"].values
    if len(lag_prod) > 0:
        if len(lag_prod) == 1:
            lag_prod_values_2.append(lag_prod[0])
        else:
            lag_prod_values_2.append(np.mean(lag_prod))
    else:
        lag_prod_values_2.append(np.nan)
        
UH_2023_TRAIN["LAG_PROD"] = lag_prod_values
UH_2023_TRAIN["LAG2_PROD"] = lag_prod_values_2

## EDA
print(DATOS_METEO.info())
print(DATOS_METEO.windDirection.unique())

msno.matrix(DATOS_METEO[[i for i in DATOS_METEO.columns if i in ["windDirection"]]])
plt.show()

print(DATOS_METEO[DATOS_METEO.pressureMeanSeaLevel.isna() == False])

# Continuamos con la transformación de los datos:

# Nos quedamos con el año, mes y día de la fecha
DATOS_METEO["Year"] = list(map(str,DATOS_METEO.validTimeUtc.dt.year))
DATOS_METEO["Month"] = list(map(str,DATOS_METEO.validTimeUtc.dt.month))
DATOS_METEO["Day"] = list(map(str,DATOS_METEO.validTimeUtc.dt.day))

# Mantenemos únicamente las columnas que consideramos importantes.
columns = ["precip1Hour", "Year", "Month", "Day", "pressureChange", "pressureMeanSeaLevel", 
          "relativeHumidity", "snow1Hour", "temperature", "temperatureDewPoint","temperatureFeelsLike",
          "uvIndex", "visibility", "windSpeed", "windDirection", "ID_ESTACION"] 

DATOS_METEO_FINAL = DATOS_METEO[[i for i in DATOS_METEO.columns if i in columns]].copy()

# Reemplazamos los valores de la columna uvIndex que marcan un NaN
DATOS_METEO_FINAL.uvIndex.replace([-2,-1], np.nan, inplace = True)

## EDA
print(DATOS_ETO.date.info())
print(list(DATOS_ETO.columns))


# Agregamos las filas de forma que obtengamos información diaria a partir de la información horaria originalmente disponible. Si nos interesa dar cierto peso a datos atípicos, usamos la media. Si no, la mediana. Finalmente, para cierto tipo de variables, mantenemos también la suma, el mínimo o el máximo, además de la moda para las variables que sean categóricas.
DATOS_METEO_FINAL = DATOS_METEO_FINAL.groupby(["ID_ESTACION", "Day", "Month", "Year"], sort = False).agg({"precip1Hour":[np.nanmean,np.nanmin,np.nanmax],
                                                                                      "relativeHumidity": np.nanmedian,
                                                                                      "snow1Hour": [np.nanmean,np.nanmin,np.nanmax], 
                                                                                      "temperature": [np.nanmedian, np.nanmax, np.nanmin],
                                                                                      })
DATOS_METEO_FINAL.columns = ["_".join(col) for col in DATOS_METEO_FINAL.columns.values]
DATOS_METEO_FINAL.columns = [i.split("_")[1][3:] + i.split("_")[0].capitalize().replace("1hour","") for i in DATOS_METEO_FINAL.columns.values] 
DATOS_METEO_FINAL = DATOS_METEO_FINAL.reset_index()

# Obtenemos el año, mes y día de los datos. Posteriormente, los transformamos para que tengan el formato adecuado.
DATOS_ETO["date"] = [i[:4] + "-" + i[4:6] + "-" + i[6:] for i in map(str,DATOS_ETO.date)]
DATOS_ETO[["Year", "Month", "Day"]] = DATOS_ETO.date.str.split("-", expand = True)
DATOS_ETO.loc[DATOS_ETO.Month.isin([i for i in DATOS_ETO.Month.unique() if int(i) <= 9]), "Month"] = [i.replace("0", "") for i in DATOS_ETO.Month if int(i) < 10]
DATOS_ETO.loc[DATOS_ETO.Day.isin([i for i in DATOS_ETO.Day.unique() if int(i) <= 9]), "Day"] = [i.replace("0", "") for i in DATOS_ETO.Day if int(i) < 10]
DATOS_ETO.drop("date", axis = 1, inplace = True)

# Juntamos los todos los datos meteorológicos en un solo dataframe:
DATOS_CLIMA = pd.merge(DATOS_METEO_FINAL, DATOS_ETO, how = "inner", on = ["ID_ESTACION", "Day", "Month", "Year"], suffixes = ("_METEO", "_ETO"))

# Mantenemos los datos de año, mes y día en el formato adecuado.
DATOS_CLIMA["Year"] = list(map(int, DATOS_CLIMA["Year"]))
DATOS_CLIMA["Month"] = list(map(int, DATOS_CLIMA["Month"]))
DATOS_CLIMA["Day"] = list(map(int, DATOS_CLIMA["Day"]))

# Obtenemos la fase del ciclo de la uva a la que pertenece cada observación, haciendo uso de la función definida al principio del script.
DATOS_CLIMA[["Phase", "PredYear"]] = pd.DataFrame(list(map(getPhase, [row for i,row in DATOS_CLIMA.iterrows()])))

# Eliminamos los datos relativos al 2015, ya que solamente tenemos dos datos, a saber, 29 y 30 de junio.
DATOS_CLIMA.drop(DATOS_CLIMA.loc[DATOS_CLIMA.PredYear == 2015].index, axis = 0, inplace = True)

print(DATOS_CLIMA.head(5))

columns = list(DATOS_CLIMA.columns)
# Mantenemos solamente las variables climáticas más importantes, relativas a la temperatura, humedad, lluvia, nieve y viento 
selected_columns = columns[:11] + columns[62:65] + columns[74:77] + columns[272:275] + columns[287:]

aggColumns = {i:np.nanmean for i in selected_columns if i not in ["ID_ESTACION", "Day", "Month", "Year", "PredYear", "Phase"]}
DATOS_CLIMA = DATOS_CLIMA.groupby(["ID_ESTACION", "PredYear", "Phase"]).agg(aggColumns).reset_index()

# Colocamos los valores en columnas con su nombre correspondiente.
values = [(est, int(float(str(year)[2:]))) for est in DATOS_CLIMA["ID_ESTACION"].unique() for year in DATOS_CLIMA["PredYear"].unique()]
data = {"ID_ESTACION":[i[0] for i in values], "CAMPAÑA":[i[1] for i in values]}
for phase in DATOS_CLIMA.Phase.unique():
    for column in [i for i in DATOS_CLIMA.columns if i not in ["ID_ESTACION", "PredYear", "Phase"]]:
        name = phase + "_" + column
        value = DATOS_CLIMA.loc[DATOS_CLIMA.Phase == phase, column].values
        data[name] = value
        
# Obtenemos el dataset meterológico final, con toda la información meteorológica en el formato adecuado.
DATOS_CLIMA_FINAL = pd.DataFrame(data = data)

cols = ["L","B","F1","F2","M","R"] # Mantenemos solamente las fases del ciclo de la vid más relevantes
DATOS_CLIMA_FINAL = DATOS_CLIMA_FINAL[[i for i in DATOS_CLIMA_FINAL.columns if True in [True for j in cols if j in i]] + ["ID_ESTACION"]]

print(DATOS_CLIMA_FINAL.head(5))

# Obtenemos el dataset final, con la información de las fincas.
WINE_DATA = pd.merge(UH_2023_TRAIN, DATOS_CLIMA_FINAL, how = "inner", on = ["ID_ESTACION", "CAMPAÑA"], suffixes = ("_TRAIN", "CLIMA")).sort_values(["ID_FINCA", "CAMPAÑA"])

# Separamos los datos de 2022 del resto.
WINE_DATA_PRED = WINE_DATA.loc[WINE_DATA.PRODUCCION.isna(), :].copy()
WINE_DATA_TRAIN = WINE_DATA.loc[WINE_DATA.PRODUCCION.isna() == False, :].copy()

# Finalmente, enviamos el ID de la finca al índice y eliminamos la columna con el año de la campaña.
WINE_DATA_PRED.set_index("ID_FINCA", inplace = True)
WINE_DATA_TRAIN.set_index("ID_FINCA", inplace = True)
WINE_DATA_PRED.drop("CAMPAÑA", axis = 1, inplace = True)
WINE_DATA_TRAIN.drop("CAMPAÑA", axis = 1, inplace = True)

# Parte de train
X = WINE_DATA_TRAIN[[i for i in WINE_DATA_TRAIN.columns if i not in ["PRODUCCION", "ID_FINCA", "ID_ZONA", "ID_ESTACION"]]]
X = X.drop(columns = X.columns[X.apply(lambda x: np.std(x) == 0, axis = 0)])  #Eliminamos todas las columnas con varianza 0
#Creamos interacciones de la superficie con la variedad de la uva y el modo de cultivo
X["SUP_VAR"] = X["SUPERFICIE"]*X["VARIEDAD"] 
X["SUP_MOD"] = X["SUPERFICIE"]*X["MODO"]
# Eliminamos aquellas columnas que no son relevantes para producir
X.drop(["L_maxPrecip", "M_medianRelativehumidity", "F2_FeelsLikeLocalDaytimeMin", "F2_FeelsLikeLocalOvernightMin", "C_medianRelativehumidity"], axis = 1, inplace = True)

# Creamos los array de los datos
X_2 = np.array(X)
y_2 = np.array(WINE_DATA_TRAIN[["PRODUCCION"]])

# Hacemos el split entre train y test
X_train_2, X_test_2, y_train_2, y_test_2 = train_test_split(X_2,y_2,random_state = 669280, shuffle = True, test_size = 0.3)


# Parte de predicción
X_pred_2 = WINE_DATA_PRED[[i for i in WINE_DATA_PRED.columns if i not in ["PRODUCCION", "ID_FINCA", "ID_ZONA", "ID_ESTACION"]]]

# Repetimos el proceso llevado a cabo en los datos de entrenamiento
X_pred_2["SUP_VAR"] = X_pred_2["SUPERFICIE"]*X_pred_2["VARIEDAD"]
X_pred_2["SUP_MOD"] = X_pred_2["SUPERFICIE"]*X_pred_2["MODO"]
X_pred_2.drop(["L_maxPrecip", "M_medianRelativehumidity", "F2_FeelsLikeLocalDaytimeMin", "F2_FeelsLikeLocalOvernightMin", "C_medianRelativehumidity"], axis = 1, inplace = True)

X_pred_2 = X_pred_2[[i for i in X_pred_2.columns if i in X.columns]]

X_pred_2 = np.array(X_pred_2)


############################# LAZY PREDICTOR #############################
###  Para obtener cuál es el mejor modelo según la métrica:
reg = LazyRegressor(verbose=0, ignore_warnings=False, custom_metric=None)
models, predictions = reg.fit(X_train_2, X_test_2, y_train_2, y_test_2)

print(models)

#Mostramos la matriz de correlación.
plt.rcParams['figure.figsize'] = [20, 15]
ax = sns.heatmap(
    corrmat, 
    vmin=-1, vmax=1, center=0,
    cmap=sns.diverging_palette(20, 220, n=200),
    square=True
)
ax.set_xticklabels(
    ax.get_xticklabels(),
    rotation=90,
    horizontalalignment='right'
)
plt.show()


############################# PREPROCESAMIENTO #############################
### CREACION DE FUNCIONES AUXILIARES PARA LA PIPELINE 

# Esta clase permite implementar funciones creadas por el usuario en una Pipeline de sklearn.
class DataframeFunctionTransformer():
    def __init__(self, func):
        self.func = func

    def transform(self, input_df, **transform_params):
        return self.func(input_df)

    def fit(self, X, y=None, **fit_params):
        return self

def efficiency_diff(dataset):
    
    """
    Esta función recibe un array de numpy con las columnas adecuadas, y devuelve el mismo array con dos columnas adicionales, una con el logaritmo de la eficiencia de la finca, que se calcula como la cantidad producida entre la superficie y otra con las primeras diferencias para la producción del año anterior al de la fila.
    """
    
    #Obtenemos las columnas necesarias para crear las variables
    superficie = dataset[:,5]
    lag_prod = dataset[:,6]
    lag2_prod = dataset[:,7]
    
    #Creamos las variables y las ponemos en la dimensión adecuada para usar np.concatenate
    diff = lag_prod - lag2_prod
    diff = diff.reshape(diff.shape[0], 1)
    eficiencia = lag_prod/superficie
    eficiencia = np.log(eficiencia)
    eficiencia = eficiencia.reshape(eficiencia.shape[0], 1)
    
    #Concatenamos las nuevas columnas
    dataset = np.concatenate((dataset, diff), axis = 1)
    dataset = np.concatenate((dataset, eficiencia), axis = 1)
    return dataset


############################# PRUEBA MODELOS #############################

# Objeto cv para el k-fold cross-validation implementado posteriormente
cv = RepeatedKFold(n_splits=10, n_repeats=3, random_state=seed_num) 

### *** XGB REGRESSOR***
# Definimos la función de búsqueda para Sequential Model Optimization (SMO):
def objective(trial):
    params = {
        #'boost' : trial.suggest_categorical(name='boost', choices=["gbtree","dart"]),
        'n_estimators': trial.suggest_int('n_estimators', 120, 1500),
        'max_depth': trial.suggest_int('max_depth', 5, 15),
        'learning_rate': trial.suggest_float('learning_rate', 0.15, 0.3),
        'colsample_bytree' : trial.suggest_float("colsample_bytree", 0.7, 1),
        #'subsample': trial.suggest_float('learning_rate', 0.8, 1),
        #'colsample_bytree': trial.suggest_float('colsample_bytree', 0.1, 1.0),
        #'min_child_weight': trial.suggest_int('min_child_weight', 1, 10),
        'random_state': seed_num,
        'eval_metric' : "rmse"
    }


    reg_tr = compose.TransformedTargetRegressor(regressor=xgb.XGBRegressor(**params, n_jobs =6), func=np.log,
                                                     inverse_func = np.exp)
    model = Pipeline([
        #("imputer", IterativeImputer(random_state=seed_num, estimator= KNeighborsRegressor(),
                                     #imputation_order = 'arabic')), #Se puede probar con algunos HP
        ("ChildColumn", DataframeFunctionTransformer(efficiency_diff)), # Obtener las columnas que queremos
        ("reg",reg_tr)])

    mean_rmse_score =  -np.mean(cross_val_score(model,X_train_2,y_train_2,cv = cv, scoring = "neg_root_mean_squared_error"
                                        , error_score = "raise", n_jobs = 6))
    return mean_rmse_score

budget = 500
study_XGB_2 = optuna.create_study(study_name = "XGBHPO", direction = "minimize")
study_XGB_2.optimize(objective, n_trials = budget)

print("The best set of parameters is: ", study_XGB_2.best_params)
print("The rmse score computed with inner-evaluation is:", study_XGB_2.best_value)
print(f"The optimal value is reached in {study_XGB_2.best_trial.number} iterations")

### *** GRADIENT BOOSTING REGRESSOR ***
# Definimos la función de búsqueda para Sequential Model Optimization (SMO):
def objective(trial):
    params = {
        'learning_rate' : trial.suggest_float("learning_rate",0.15, 0.3),
        'n_estimators' : trial.suggest_int("max_iter",50,1500),
        'max_depth' : trial.suggest_int("max_depth",5,15),
        'random_state': seed_num,
        'loss' : "squared_error"
    }

    model = Pipeline([
        ("imputer", IterativeImputer()), #Se puede probar con algunos HP
        ("ChildColumn", DataframeFunctionTransformer(efficiency_diff)), # Obtener las columnas que queremos
        ("reg",GradientBoostingRegressor(**params))])

    mean_rmse_score =  -np.mean(cross_val_score(model,X_train_2,y_train_2,cv = cv, scoring = "neg_root_mean_squared_error"
                                        , error_score = "raise", n_jobs = 6))

    return mean_rmse_score

budget = 300
study_GB_2 = optuna.create_study(study_name = "GBHPO", direction = "minimize")
study_GB_2.optimize(objective, n_trials = budget)

# Inner evaluation
print("The best set of parameters is: ", study_GB_2.best_params)
print("The rmse score computed with inner-evaluation is:", study_GB_2.best_value)
print(f"The optimal value is reached in {study_GB_2.best_trial.number} iterations")

### *** LGBM REGRESSOR ***
# Definimos la función de búsqueda para Sequential Model Optimization (SMO):
def objective(trial):

    params = {
        'n_estimators' : trial.suggest_int("n_estimators", 120, 1500),
        'num_leaves' : trial.suggest_int("num_leaves", 20, 40),
        'max_depth' :  trial.suggest_int("max_depth", 5,15),
        'learning_rate' : trial.suggest_float("learning_rate", 1e-2, 0.2),
        'random_state': seed_num,
        'eval_metric' : "rmse"
    }

    reg_tr = compose.TransformedTargetRegressor(regressor=LGBMRegressor(**params, n_jobs =6), func=np.log,
                                                     inverse_func = np.exp)
    model = Pipeline([
        ("imputer", IterativeImputer()), #Se puede probar con algunos HP
        ("ChildColumn", DataframeFunctionTransformer(efficiency_diff)),
        ("reg",reg_tr)])
    mean_rmse_score =  -np.mean(cross_val_score(model,X_train_2,y_train_2,cv = cv, scoring = "neg_root_mean_squared_error"
                                        , error_score = "raise", n_jobs = 6))

    return mean_rmse_score

budget = 500
study_LGBM_2 = optuna.create_study(study_name = "HGBHPO", direction = "minimize")
study_LGBM_2.optimize(objective, n_trials = budget)

# Inner evaluation
print("The best set of parameters is: ", study_LGBM_2.best_params)
print("The rmse score computed with inner-evaluation is:", study_LGBM_2.best_value)
print(f"The optimal value is reached in {study_LGBM_2.best_trial.number} iterations")
    
### *** HIST GRADIENT BOOSTING (MODELO ESCOGIDO) ***
# Definimos la función de búsqueda para Sequential Model Optimization (SMO):
def objective(trial):
    params = {
        'learning_rate' : trial.suggest_float("learning_rate",0.15, 0.3),
        'max_iter' : trial.suggest_int("max_iter",50,1500),
        'max_leaf_nodes' : trial.suggest_int("max_leaf_nodes",10,200),
        'max_depth' : trial.suggest_int("max_depth",5,15),
        'random_state': seed_num,
        'loss' : "squared_error"
    }

    model = Pipeline([
        ("ChildColumn", DataframeFunctionTransformer(efficiency_diff)), 
        ("reg",HistGradientBoostingRegressor(**params))])

    mean_rmse_score =  -np.mean(cross_val_score(model,X_train_2,y_train_2.ravel(),cv = cv, scoring = "neg_root_mean_squared_error"
                                        , error_score = "raise", n_jobs = 6))

    return mean_rmse_score

budget = 500
study_HGB_2 = optuna.create_study(study_name = "HGBHPO", direction = "minimize")
study_HGB_2.optimize(objective, n_trials = budget)

# Inner evaluation
print("The best set of parameters is: ", study_HGB_2.best_params)
print("The rmse score computed with inner-evaluation is:", study_HGB_2.best_value)
print(f"The optimal value is reached in {study_HGB_2.best_trial.number} iterations")

# Como ha sido la mejor inner, hacemos la outer evaluation
reg_tr = compose.TransformedTargetRegressor(regressor=xgb.XGBRegressor(**study_XGB_2.best_params, n_jobs =6), func=np.log,
                                                     inverse_func = np.exp)
model_XGB_2 = Pipeline([
        #("imputer", IterativeImputer(random_state=seed_num, estimator= KNeighborsRegressor(),
                                    # imputation_order = 'arabic')), #Se puede probar con algunos HP
        ("ChildColumn", DataframeFunctionTransformer(efficiency_diff)), # Obtener las columnas que queremos
        ("reg",reg_tr)])
model_XGB_2.fit(X_train_2,y_train_2)
y_pred_2 = model_XGB_2.predict(X_test_2)
score = mean_squared_error(y_test_2, y_pred_2, squared = False)
print("The RMSE in the test data is: ", score)

# Definimos el modelo
best_params_HGB = {'learning_rate': 0.16966582726087054, 'max_iter': 50, 'max_leaf_nodes': 14, 'max_depth': 6}
model_HGB_2 = Pipeline([
        ("imputer", IterativeImputer()),
        ("ChildColumn", DataframeFunctionTransformer(efficiency_diff)), # Obtener las columnas que queremos
        ("reg",HistGradientBoostingRegressor(**best_params_HGB))])
# Entrenamos
model_HGB_2.fit(X_2,y_2.ravel())

# Predecimos
y_pred_2 = model_HGB_2.predict(X_pred_2)

exp2 = dx.Explainer(model_HGB_2, X_2, np.squeeze(y_2))
mp2 = exp2.model_parts(loss_function = 'rmse')

# we can plot the results
mp2.plot()

relevant_var_idx2 = list(mp2.result.variable[-11:-1])
relevant_cols2 = [X.columns[int(idx)] for idx in relevant_var_idx2]
relevant_cols2.reverse()

print("The most relevant variables for the analysis of HGB are: ")
print(relevant_cols2)

sample_data = X_2[100,:]
rf_pparts = exp2.predict_parts(new_observation = sample_data, type = "break_down")
rf_pparts.plot()


protected_21 = np.where(X_2[:,2] == 1, "modo_1", "else")
protected_22 = np.where(X_2[:,2] == 2, "modo_2", "else")
protected_30 = np.where(X_2[:,3] == 0, "tipo_0", "else")
protected_31 = np.where(X_2[:,3] == 1, "tipo_1", "else")
protected_40 = np.where(X_2[:,4] == 0, "color_0", "else")
protected_41 = np.where(X_2[:,4] == 1, "color_1", "else")
privileged = "else"

fobject2 = exp2.model_fairness(protected_21, privileged)
fobject2.fairness_check()
fobject2 = exp2.model_fairness(protected_22, privileged)
fobject2.fairness_check()
fobject2 = exp2.model_fairness(protected_30, privileged)
fobject2.fairness_check()
fobject2 = exp2.model_fairness(protected_31, privileged)
fobject2.fairness_check()
fobject2 = exp2.model_fairness(protected_40, privileged)
fobject2.fairness_check()
fobject2 = exp2.model_fairness(protected_41, privileged)
fobject2.fairness_check()

# Ajustamos el formato
y_pred_2 = y_pred_2.flatten()
pred_data = WINE_DATA_PRED.reset_index()[["ID_FINCA","VARIEDAD","MODO","TIPO","COLOR","SUPERFICIE"]].copy()
pred_data["PRODUCCION"] = list(map(lambda x: round(x, 2), y_pred_2))
pred_data = pred_data.sort_values(["ID_FINCA","VARIEDAD","MODO","TIPO","COLOR","SUPERFICIE","PRODUCCION"])

#Obtenemos el .txt final.
pred_data.to_csv("E-No-Logos.txt", header = False, sep='|', index = False)

### EL SIGUIENTE CÓDIGO REPITE EL MISMO PROCESO, PERO PARA OTRO TIPO DE AGRUPACIÓN DE LOS DATOS.
## Como no se emplea en el modelo final, se deja comentado.
"""
import matplotlib.pyplot as plt
plt.style.use("ggplot")
import pandas as pd
import numpy as np
import missingno as msno
from scipy import stats
import math
from sklearn.impute import KNNImputer, SimpleImputer
from sklearn.pipeline import Pipeline, make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split
from sklearn.model_selection import cross_val_score, KFold, RepeatedKFold
from sklearn.metrics import mean_squared_error
from sklearn.feature_selection import SelectKBest, f_regression, mutual_info_regression
import optuna
import numpy as np
from lazypredict.Supervised import LazyRegressor
from lightgbm import LGBMRegressor
import xgboost as xgb
import seaborn as sns
from sklearn.preprocessing import PolynomialFeatures
import pathlib
from sklearn.ensemble import ExtraTreesRegressor, RandomForestRegressor, GradientBoostingRegressor, StackingRegressor, HistGradientBoostingRegressor
from sklearn.linear_model import LinearRegression
import random

seed_num = 669280

# Los datos se encuentran en la carpeta actual, si no, hay que poner ahí el path de la carpeta con los datos
current_folder = str(pathlib.Path().resolve())

# Datos de las viñas:
UH_2023_TRAIN = pd.read_csv((current_folder+"\\UH_2023_TRAIN.txt")
                            , sep = "|", encoding = "utf-8")

#Datos de la información meteorológica sin agregar.    
DATOS_METEO = pd.read_csv((current_folder+"\\DATOS_METEO.txt")
                            , sep = "|", encoding = "utf-8", parse_dates = [0])

# Datos meteorológicos agregados:
DATOS_ETO = pd.read_csv((current_folder+"\\DATOS_ETO.txt")
                            , sep = "|", encoding = "utf-8")
                            
def getPhase(x):
    
    Esta función redibe una fila (x) de un dataframe de pandas y devuelve la
    fase del ciclo de la viña a la que pertenece, en función de la fecha en la
    que se recogieron los datos. Finalmente, si la fecha es superior al 30 de junio de 
    ese año, señala que esos datos deberían utilzarse en la producción del año siguiente.
    
    day = x["Day"]
    month = x["Month"]
    year = x["Year"]
    if day < 10:
        day = "0" + str(day).split(".")[0]
    else:
        day = str(day).split(".")[0]
        
    month = str(month).split(".")[0]
    
    key = int(month + day)
    
    if key >= 301 and key <= 310:
        return "Mar1", int(year)
    elif key >= 311 and key <= 320:
        return "Mar2", int(year)
    elif key >= 321 and key <= 330:
        return "Mar3", int(year)
    elif key >= 401 and key <= 410:
        return "Abr1", int(year)
    elif key >= 411 and key <= 420:
        return "Abr2", int(year)
    elif key >= 421 and key <= 430:
        return "Abr3", int(year)
    elif key >= 501 and key <= 510:
        return "May1", int(year)
    elif key >= 511 and key <= 520:
        return "May2", int(year)
    elif key >= 521 and key <= 530:
        return "May3", int(year)
    elif key >= 601 and key <= 610:
        return "Jun1", int(year)
    elif key >= 611 and key <= 620:
        return "Jun2", int(year)
    elif key >= 621 and key <= 630:
        return "Jun3", int(year)
    elif key >= 701 and key <= 710:
        year+= 1
        return "Jul1", int(year)
    elif key >= 711 and key <= 720:
        year+= 1
        return "Jul2", int(year)
    elif key >= 721 and key <= 730:
        year+= 1
        return "Jul3", int(year)
    elif key >= 801 and key <= 810:
        year+= 1
        return "Ago1", int(year)
    elif key >= 811 and key <= 820:
        year+= 1
        return "Ago2", int(year)
    elif key >= 821 and key <= 830:
        year+= 1
        return "Ago3", int(year)
    elif key >= 901 and key <= 910:
        year += 1
        return "Sep1", int(year)
    elif key >= 911 and key <= 920:
        year += 1
        return "Sep2", int(year)
    elif key >= 921 and key <= 930:
        year += 1
        return "Sep3", int(year)
    elif key >= 1001 and key <= 1010:
        year += 1
        return "Oct1", int(year)
    elif key >= 1011 and key <= 1020:
        year += 1
        return "Oct2", int(year)
    elif key >= 1021 and key <= 1030:
        year += 1
        return "Oct3", int(year)
    elif key >= 1101 and key <= 1110:
        year += 1
        return "Nov1", int(year)
    elif key >= 1111 and key <= 1120:
        year += 1
        return "Nov2", int(year)
    elif key >= 1121 and key <= 1130:
        year += 1
        return "Nov3", int(year)
    elif key >= 1201 and key <= 1210:
        year += 1
        return "Dic1", int(year)
    elif key >= 1211 and key <= 1220:
        year += 1
        return "Dic2", int(year)
    else:
        if key >= 1221 and key <= 1231:
            year += 1
            return "Dic3", int(year)
    
def getHeight(x):
    
    Esta función recibe una fila (x) de un dataframe de pandas, y devuelve los datos de la columna ALTITUD en el
    formato adecuado para ajustar el modelo y realizar las predicciones.
    
    if type(x["ALTITUD"]) == str:
        alt = [float(i) for i in x["ALTITUD"].split("-")]
        if len(alt) > 2:
            return (alt[0] + alt[1])/2
        else:
            return alt[0]
    else:
        return x["ALTITUD"]
    

# Datos de las viñas
## EDA 
print(UH_2023_TRAIN.info())

print(UH_2023_TRAIN.sort_values(["ID_FINCA","CAMPAÑA"]).head(20))
print(UH_2023_TRAIN.CAMPAÑA.unique())

print(UH_2023_TRAIN[[i for i in UH_2023_TRAIN if i not in ["CAMPAÑA", "ID_FINCA", "ID_ZONA", "ID_ESTACION"]]].describe())

## Transformamos los datos
UH_2023_TRAIN.SUPERFICIE.replace(0.0, np.nan, inplace = True)
UH_2023_TRAIN.ALTITUD = list(map(getHeight, [row for i,row in UH_2023_TRAIN.iterrows()]))

# Información sobre la superficie de las fincas en base a diferentes variables de uva:
INFO_SUPERFICIE = UH_2023_TRAIN[["ID_FINCA", "VARIEDAD", "MODO", "TIPO", "SUPERFICIE"]].sort_values(["ID_FINCA","VARIEDAD", "MODO", "TIPO"]).dropna().drop_duplicates().reset_index().drop("index", axis = 1)

# Imputando los valores de la superficie, en base a diferentes variables de la uva, si están disponibles:

todo = 0 #Número de fincas imputadas que mantuvieron Variedad, Modo y Tipo de la uva.
part1 = 0 #Número de fincas imputadas que mantuvieron Variedad y Modo de la uva.
part2 = 0 #Número de fincas imputadas que mantuvieron Variedad de la uva.
part3 = 0 #Número de fincas imputadas que no mantuvieron ninguna variable de la uva.

for i,row in UH_2023_TRAIN.iterrows():
    if math.isnan(row["SUPERFICIE"]):
        sup_value = list(INFO_SUPERFICIE.loc[(INFO_SUPERFICIE.ID_FINCA == row["ID_FINCA"]) & (INFO_SUPERFICIE.VARIEDAD == row["VARIEDAD"]) & (INFO_SUPERFICIE.MODO == row["MODO"]) & (INFO_SUPERFICIE.TIPO == row["TIPO"]), "SUPERFICIE"])
        if len(sup_value) > 0:
            todo+=1
            if len(sup_value) > 1:
                UH_2023_TRAIN.loc[i, "SUPERFICIE"] = np.mean(sup_value) #Si hay varias coincidencias, imputamos la media de la superficie
            else:
                UH_2023_TRAIN.loc[i, "SUPERFICIE"] = sup_value[0]
        else:
            sup_value = list(INFO_SUPERFICIE.loc[(INFO_SUPERFICIE.ID_FINCA == row["ID_FINCA"]) & (INFO_SUPERFICIE.VARIEDAD == row["VARIEDAD"]) & (INFO_SUPERFICIE.MODO == row["MODO"]), "SUPERFICIE"])
            if len(sup_value) > 0:
                part1+=1
                if len(sup_value) > 1:
                    UH_2023_TRAIN.loc[i, "SUPERFICIE"] = np.mean(sup_value)
                else:
                    UH_2023_TRAIN.loc[i, "SUPERFICIE"] = sup_value[0]
            else:
                sup_value = list(INFO_SUPERFICIE.loc[(INFO_SUPERFICIE.ID_FINCA == row["ID_FINCA"]) & (INFO_SUPERFICIE.VARIEDAD == row["VARIEDAD"]), "SUPERFICIE"])
                if len(sup_value) > 0:
                    part2+=1
                    if len(sup_value) > 1:
                        UH_2023_TRAIN.loc[i, "SUPERFICIE"] = np.mean(sup_value)
                    else:
                        UH_2023_TRAIN.loc[i, "SUPERFICIE"] = sup_value[0]
                else:
                    sup_value = list(INFO_SUPERFICIE.loc[(INFO_SUPERFICIE.ID_FINCA == row["ID_FINCA"]), "SUPERFICIE"])
                    if len(sup_value) > 0:
                        part3+=1
                        if len(sup_value) > 1:
                            UH_2023_TRAIN.loc[i, "SUPERFICIE"] = np.mean(sup_value)
                        else:
                            UH_2023_TRAIN.loc[i, "SUPERFICIE"] = sup_value[0]
                    else:
                        UH_2023_TRAIN.loc[i, "SUPERFICIE"] = np.nan #Si no tenemos datos para la superficie de la finca, imputamos un NaN.
    else:
        pass
    
#Datos de la información meteorológica sin agregar
## EDA
print(DATOS_METEO.info())
print(DATOS_METEO.windDirection.unique())

msno.matrix(DATOS_METEO[[i for i in DATOS_METEO.columns if i in ["windDirection"]]])
plt.show()

print(DATOS_METEO[DATOS_METEO.pressureMeanSeaLevel.isna() == False])

#Continuamos con la transformación de los datos:

#Nos quedamos con el año, mes y día de la fecha
DATOS_METEO["Year"] = list(map(str,DATOS_METEO.validTimeUtc.dt.year))
DATOS_METEO["Month"] = list(map(str,DATOS_METEO.validTimeUtc.dt.month))
DATOS_METEO["Day"] = list(map(str,DATOS_METEO.validTimeUtc.dt.day))

#Mantenemos únicamente las columnas que consideramos importantes.
columns = ["precip1Hour", "Year", "Month", "Day", "pressureChange", "pressureMeanSeaLevel", 
          "relativeHumidity", "snow1Hour", "temperature", "temperatureDewPoint","temperatureFeelsLike",
          "uvIndex", "visibility", "windSpeed", "windDirection", "ID_ESTACION"] 

DATOS_METEO_FINAL = DATOS_METEO[[i for i in DATOS_METEO.columns if i in columns]].copy()

#Reemplazamos los valores de la columna uvIndex que marcan un NaN
DATOS_METEO_FINAL.uvIndex.replace([-2,-1], np.nan, inplace = True)


#Agregamos las filas de forma que obtengamos información diaria a partir de la información horaria originalmente disponible. Si nos interesa dar cierto peso a datos atípicos, usamos la media. Si no, la mediana. Finalmente, para cierto tipo de variables, mantenemos también la suma, el mínimo o el máximo, además de la moda para las variables que sean categóricas.
DATOS_METEO_FINAL = DATOS_METEO_FINAL.groupby(["ID_ESTACION", "Day", "Month", "Year"], sort = False).agg({"precip1Hour":[np.nanmean,np.nanmin,np.nanmax],
                                                                                      "relativeHumidity": np.nanmedian,
                                                                                      "snow1Hour": [np.nanmean,np.nanmin,np.nanmax], 
                                                                                      "temperature": [np.nanmedian, np.nanmax, np.nanmin],
                                                                                      })
DATOS_METEO_FINAL.columns = ["_".join(col) for col in DATOS_METEO_FINAL.columns.values]
DATOS_METEO_FINAL.columns = [i.split("_")[1][3:] + i.split("_")[0].capitalize().replace("1hour","") for i in DATOS_METEO_FINAL.columns.values] 
DATOS_METEO_FINAL = DATOS_METEO_FINAL.reset_index()

# Datos meteorológicos agregados
## EDA
print(DATOS_ETO.date.info())
print(list(DATOS_ETO.columns))

#Obtenemos el año, mes y día de los datos. Posteriormente, los transformamos para que tengan el formato adecuado.
DATOS_ETO["date"] = [i[:4] + "-" + i[4:6] + "-" + i[6:] for i in map(str,DATOS_ETO.date)]
DATOS_ETO[["Year", "Month", "Day"]] = DATOS_ETO.date.str.split("-", expand = True)
DATOS_ETO.loc[DATOS_ETO.Month.isin([i for i in DATOS_ETO.Month.unique() if int(i) <= 9]), "Month"] = [i.replace("0", "") for i in DATOS_ETO.Month if int(i) < 10]
DATOS_ETO.loc[DATOS_ETO.Day.isin([i for i in DATOS_ETO.Day.unique() if int(i) <= 9]), "Day"] = [i.replace("0", "") for i in DATOS_ETO.Day if int(i) < 10]
DATOS_ETO.drop("date", axis = 1, inplace = True)

# Juntamos los todos los datos meteorológicos en un solo dataframe:
DATOS_CLIMA = pd.merge(DATOS_METEO_FINAL, DATOS_ETO, how = "inner", on = ["ID_ESTACION", "Day", "Month", "Year"], suffixes = ("_METEO", "_ETO"))

#Mantenemos los datos de año, mes y día en el formato adecuado.
DATOS_CLIMA["Year"] = list(map(int, DATOS_CLIMA["Year"]))
DATOS_CLIMA["Month"] = list(map(int, DATOS_CLIMA["Month"]))
DATOS_CLIMA["Day"] = list(map(int, DATOS_CLIMA["Day"]))

#Obtenemos la fase del ciclo de la uva a la que pertenece cada observación, haciendo uso de la función definida al principio del script.
DATOS_CLIMA[["Phase", "PredYear"]] = pd.DataFrame(list(map(getPhase, [row for i,row in DATOS_CLIMA.iterrows()])))

#Eliminamos los datos relativos al 2015, ya que solamente tenemos dos datos, a saber, 29 y 30 de junio.
DATOS_CLIMA.drop(DATOS_CLIMA.loc[DATOS_CLIMA.PredYear == 2015].index, axis = 0, inplace = True)

display(DATOS_CLIMA)

columns = list(DATOS_CLIMA.columns)
selected_columns = columns[:11] + columns[62:65] + columns[74:77] + columns[287:]

aggColumns = {i:np.nanmean for i in selected_columns if i not in ["ID_ESTACION", "Day", "Month", "Year", "PredYear", "Phase"]}
DATOS_CLIMA = DATOS_CLIMA.groupby(["ID_ESTACION", "PredYear", "Phase"]).agg(aggColumns).reset_index()
print(DATOS_CLIMA.head())

#Colocamos los valores en columnas con su nombre correspondiente.
values = [(est, int(float(str(year)[2:]))) for est in DATOS_CLIMA["ID_ESTACION"].unique() for year in DATOS_CLIMA["PredYear"].unique()]
data = {"ID_ESTACION":[i[0] for i in values], "CAMPAÑA":[i[1] for i in values]}
for phase in DATOS_CLIMA.Phase.unique():
    for column in [i for i in DATOS_CLIMA.columns if i not in ["ID_ESTACION", "PredYear", "Phase"]]:
        name = phase + "_" + column
        value = DATOS_CLIMA.loc[DATOS_CLIMA.Phase == phase, column].values
        data[name] = value
        
#Obtenemos el dataset meterológico final, con toda la información meteorológica en el formato adecuado.
DATOS_CLIMA_FINAL = pd.DataFrame(data = data)

cols = ["Mar1","Mar2", "Mar3", "Apr1", "Apr2", "Apr3","Jun1", "Jun2", "Jun3", "Sep1", "Dec1"]
DATOS_CLIMA_FINAL = DATOS_CLIMA_FINAL[[i for i in DATOS_CLIMA_FINAL.columns if True in [True for j in cols if j in i]] + ["ID_ESTACION", "CAMPAÑA","Ago1_FeelsLikeLocalDaytimeMax"
        ,"Jul1_FeelsLikeLocalDaytimeMax", "Jul3_FeelsLikeLocalDaytimeMax", "May1_meanPrecip", "May2_meanPrecip", "May3_meanPrecip"]]

print(DATOS_CLIMA_FINAL.head())

# Obtenemos el dataset final, con la información de las fincas.
WINE_DATA = pd.merge(UH_2023_TRAIN, DATOS_CLIMA_FINAL, how = "inner", on = ["ID_ESTACION", "CAMPAÑA"], suffixes = ("_TRAIN", "CLIMA")).sort_values(["ID_FINCA", "CAMPAÑA"])

#Separamos los datos de 2022 del resto.
WINE_DATA_PRED = WINE_DATA.loc[WINE_DATA.PRODUCCION.isna(), :].copy()
WINE_DATA_TRAIN = WINE_DATA.loc[WINE_DATA.PRODUCCION.isna() == False, :].copy()

#Finalmente, enviamos el ID de la finca al índice y eliminamos la columna con el año de la campaña.
WINE_DATA_PRED.set_index("ID_FINCA", inplace = True)
WINE_DATA_TRAIN.set_index("ID_FINCA", inplace = True)
WINE_DATA_PRED.drop("CAMPAÑA", axis = 1, inplace = True)
WINE_DATA_TRAIN.drop("CAMPAÑA", axis = 1, inplace = True)


na_values = WINE_DATA_TRAIN[WINE_DATA_TRAIN.columns[WINE_DATA_TRAIN.isna().any()]].isna().sum().sort_values(ascending = False).to_frame().reset_index()
na_values.columns = ["Columns","NaValues"]
def obtainColumns(x):
    colName = x["Columns"]
    if "_" in colName:
        colName = colName[colName.find("_")+1:]
        if "Local" in colName: 
            return colName[:colName.find("Local")]
        else:
            return colName
    else:
        return colName
    
columns = na_values["Columns"]

na_values["Columns"] = list(map(obtainColumns, [row for i,row in na_values.iterrows()]))
na_values.drop_duplicates(inplace = True)

cols = list(na_values.Columns.unique())
cols.remove("ALTITUD")
cols.remove("SUPERFICIE")

WINE_DATA_TRAIN = WINE_DATA_TRAIN[[i for i in WINE_DATA_TRAIN.columns if True not in [True for j in cols if j in i]]]
WINE_DATA_PRED = WINE_DATA_PRED[[i for i in WINE_DATA_PRED.columns if True not in [True for j in cols if j in i]]]


# Datos train
WINE_DATA_TRAIN.dropna(inplace = True)
print(WINE_DATA_TRAIN.shape)

X = WINE_DATA_TRAIN[[i for i in WINE_DATA_TRAIN.columns if i not in ["PRODUCCION", "ID_FINCA", "ID_ZONA", "ID_ESTACION"]]]
var_cls =  X.columns[X.apply(lambda x: np.std(x) == 0, axis = 0)]
X = X.drop(columns = X.columns[X.apply(lambda x: np.std(x) == 0, axis = 0)])
X["SUP_VAR"] = X["SUPERFICIE"]*X["VARIEDAD"]
X["SUP_MOD"] = X["SUPERFICIE"]*X["MODO"]
X.drop(["Mar2_medianRelativehumidity", "Jun2_maxPrecip", "Mar3_FeelsLikeLocalOvernightMin"], axis = 1, inplace = True)
corrmat = X.corr()
data = X.copy()
cols = X.columns
X_1 = np.array(X)
print(X.shape)
y_1 = np.array(WINE_DATA_TRAIN[["PRODUCCION"]])

X_train_1, X_test_1, y_train_1, y_test_1 = train_test_split(X_1,y_1,random_state = 669280, shuffle = True, test_size = 0.3)

# Datos pred
print(WINE_DATA_PRED.shape)
X_pred_1 = WINE_DATA_PRED[[i for i in WINE_DATA_PRED.columns if i not in ["PRODUCCION", "ID_FINCA", "ID_ZONA", "ID_ESTACION"]]]

# Hay NAs en altitud y superficie
cols_with_na = X_pred_1.isna().any()
num_cols_with_na = cols_with_na.sum()
print(f"Columnas con valores faltantes: {cols_with_na}")
print(f"Número de columnas con valores faltantes: {num_cols_with_na}")

# Imputación en superficie
sup_med = np.nanmedian(X_pred_1['SUPERFICIE'])
X_pred_1['SUPERFICIE'].fillna(sup_med, inplace = True)

# Imputación en altitud
alt_med = np.nanmedian(X_pred_1['ALTITUD'])
X_pred_1['ALTITUD'].fillna(alt_med, inplace = True)

X_pred_1.dropna(inplace = True) # Ahora ya no hace nada
#var_cls =  X_pred_1.columns[X_pred_1.apply(lambda x: np.std(x) == 0, axis = 0)]
#X_pred_1 = X_pred_1.drop(columns = X_pred_1.columns[X_pred_1.apply(lambda x: np.std(x) == 0, axis = 0)])
X_pred_1["SUP_VAR"] = X_pred_1["SUPERFICIE"]*X_pred_1["VARIEDAD"]
X_pred_1["SUP_MOD"] = X_pred_1["SUPERFICIE"]*X_pred_1["MODO"]
X_pred_1.drop(["Mar2_medianRelativehumidity", "Jun2_maxPrecip", "Mar3_FeelsLikeLocalOvernightMin"], axis = 1, inplace = True)
X_pred_1 = X_pred_1[[i for i in X_pred_1.columns if i in cols]]
X_pred_1 = np.array(X_pred_1)
print(X_pred_1.shape)

#Mostramos la matriz de correlación.
plt.rcParams['figure.figsize'] = [20, 15]
ax = sns.heatmap(
    corrmat, 
    vmin=-1, vmax=1, center=0,
    cmap=sns.diverging_palette(20, 220, n=200),
    square=True
)
ax.set_xticklabels(
    ax.get_xticklabels(),
    rotation=90,
    horizontalalignment='right'
)
plt.show()

cv = RepeatedKFold(n_splits=10, n_repeats=3, random_state=seed_num)

def objective(trial):
    # Define the search space for hyperparameters
    params = {
        #'boost' : trial.suggest_categorical(name='boost', choices=["gbtree","dart"]),
        'n_estimators': trial.suggest_int('n_estimators', 120, 1500),
        'max_depth': trial.suggest_int('max_depth', 5, 10),
        'learning_rate': trial.suggest_float('learning_rate', 0.15, 0.3),
        'colsample_bytree' : trial.suggest_float("colsample_bytree", 0.7, 1),
        #'subsample': trial.suggest_float('learning_rate', 0.8, 1),
        #'colsample_bytree': trial.suggest_float('colsample_bytree', 0.1, 1.0),
        #'min_child_weight': trial.suggest_int('min_child_weight', 1, 10),
        'random_state': seed_num,
        'eval_metric' : "rmse"
    }


    # Define the model
    model = Pipeline([
        #("scaler", StandardScaler()), # No es necesario escalar con algoritmos basados en árboles
        ("reg",xgb.XGBRegressor(**params, n_jobs =6))])

    # Calculate the mean RMSE score across all cross-validation folds
    mean_rmse_score =  -np.mean(cross_val_score(model,X_1,y_1,cv = cv, scoring = "neg_root_mean_squared_error"
                                        , error_score = "raise", n_jobs = 6))

    return mean_rmse_score
    
budget = 100
study_XGB = optuna.create_study(study_name = "XGBHPO", direction = "minimize")
study_XGB.optimize(objective, n_trials = budget)

print("The best set of parameters is: ", study_XGB.best_params)
print("The rmse score computed with inner-evaluation is:", study_XGB.best_value)
print(f"The optimal value is reached in {study_XGB.best_trial.number} iterations")
"""