### SCRIPT PREDICCIÓN
# Autores: Gabriel Abilleira, Paula Álvarez y Javier Godos 
# Equipo: E-No_Logos (UC3M)

### IMPORTANDO LOS PAQUETES
import matplotlib.pyplot as plt
plt.style.use("ggplot")
import pandas as pd
import numpy as np
import missingno as msno
from scipy import stats
import math
from sklearn.experimental import enable_iterative_imputer
from sklearn.impute import KNNImputer, SimpleImputer, IterativeImputer
from sklearn.pipeline import Pipeline, make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split
from sklearn.model_selection import cross_val_score, KFold, RepeatedKFold
from sklearn.metrics import mean_squared_error
from sklearn.feature_selection import SelectKBest, f_regression, mutual_info_regression
import optuna
import numpy as np
from lazypredict.Supervised import LazyRegressor
from lightgbm import LGBMRegressor
import xgboost as xgb
import seaborn as sns
from sklearn.preprocessing import PolynomialFeatures
import pathlib
from sklearn.ensemble import ExtraTreesRegressor, RandomForestRegressor, GradientBoostingRegressor, StackingRegressor, HistGradientBoostingRegressor
from sklearn.linear_model import LinearRegression, PassiveAggressiveRegressor
import random
from sklearn import compose
from sklearn.neighbors import KNeighborsRegressor
import dalex as dx

#Definición de la semilla para reproducibilidad

seed_num = 669280
## **Obtención DF**

#Cargamos los datos

# Los datos se encuentran en la carpeta actual, si no, hay que poner ahí el path de la carpeta con los datos
current_folder = str(pathlib.Path().resolve())

# Datos de las viñas:
UH_2023_TRAIN = pd.read_csv((current_folder+"\\UH_2023_TRAIN.txt")
                            , sep = "|", encoding = "utf-8")

#Datos de la información meteorológica sin agregar.    
DATOS_METEO = pd.read_csv((current_folder+"\\DATOS_METEO.txt")
                            , sep = "|", encoding = "utf-8", parse_dates = [0])

# Datos meteorológicos agregados:
DATOS_ETO = pd.read_csv((current_folder+"\\DATOS_ETO.txt")
                            , sep = "|", encoding = "utf-8")

### DEFINICIÓN DE FUNCIONES
def getPhase(x):
    
    """
    Esta función redibe una fila (x) de un dataframe de pandas y devuelve la
    fase del ciclo de la viña a la que pertenece, en función de la fecha en la
    que se recogieron los datos. Finalmente, si la fecha es superior al 30 de junio de 
    ese año, señala que esos datos deberían utilzarse en la producción del año siguiente.
    """
    
    day = x["Day"]
    month = x["Month"]
    year = x["Year"]
    if day < 10:
        day = "0" + str(day).split(".")[0]
    else:
        day = str(day).split(".")[0]
        
    month = str(month).split(".")[0]
    
    key = int(month + day)
    
    if key >= 305 and key <= 323:
        return "L", int(year)
    elif key >= 324 and key <= 520:
        return "B", int(year)
    elif key >= 521 and key <= 630:
        return "F1", int(year)
    elif key >= 701 and key <= 810:
        year+= 1
        return "F2", int(year)
    elif key >= 811 and key <= 931:
        year += 1
        return "M", int(year)
    elif key >= 1001 and key <= 1115:
        year += 1
        return "C", int(year)
    else:
        if key >= 1116 and key <= 1231:
            year += 1
        return "R", int(year)
    
    
def getHeight(x):
    
    """
    Esta función recibe una fila (x) de un dataframe de pandas, y devuelve los datos de la columna ALTITUD en el
    formato adecuado para ajustar el modelo y realizar las predicciones.
    """
    
    if type(x["ALTITUD"]) == str:
        alt = [float(i) for i in x["ALTITUD"].split("-")]
        if len(alt) > 2:
            return (alt[0] + alt[1])/2
        else:
            return alt[0]
    else:
        return x["ALTITUD"]

### CARGANDO Y TRANSFORMANDO LOS DATASETS

## Transformamos los datos
UH_2023_TRAIN.SUPERFICIE.replace(0.0, np.nan, inplace = True)
UH_2023_TRAIN.ALTITUD = list(map(getHeight, [row for i,row in UH_2023_TRAIN.iterrows()]))

# Información sobre la superficie de las fincas en base a diferentes variables de uva:
INFO_SUPERFICIE = UH_2023_TRAIN[["ID_FINCA", "VARIEDAD", "MODO", "TIPO", "SUPERFICIE"]].sort_values(["ID_FINCA","VARIEDAD", "MODO", "TIPO"]).dropna().drop_duplicates().reset_index().drop("index", axis = 1)

# Imputando los valores de la superficie, en base a diferentes variables de la uva, si están disponibles:

todo = 0 #Número de fincas imputadas que mantuvieron Variedad, Modo y Tipo de la uva.
part1 = 0 #Número de fincas imputadas que mantuvieron Variedad y Modo de la uva.
part2 = 0 #Número de fincas imputadas que mantuvieron Variedad de la uva.
part3 = 0 #Número de fincas imputadas que no mantuvieron ninguna variable de la uva.

for i,row in UH_2023_TRAIN.iterrows():
    if math.isnan(row["SUPERFICIE"]):
        sup_value = list(INFO_SUPERFICIE.loc[(INFO_SUPERFICIE.ID_FINCA == row["ID_FINCA"]) & (INFO_SUPERFICIE.VARIEDAD == row["VARIEDAD"]) & (INFO_SUPERFICIE.MODO == row["MODO"]) & (INFO_SUPERFICIE.TIPO == row["TIPO"]), "SUPERFICIE"])
        if len(sup_value) > 0:
            todo+=1
            if len(sup_value) > 1:
                UH_2023_TRAIN.loc[i, "SUPERFICIE"] = np.mean(sup_value) #Si hay varias coincidencias, imputamos la media de la superficie
            else:
                UH_2023_TRAIN.loc[i, "SUPERFICIE"] = sup_value[0]
        else:
            sup_value = list(INFO_SUPERFICIE.loc[(INFO_SUPERFICIE.ID_FINCA == row["ID_FINCA"]) & (INFO_SUPERFICIE.VARIEDAD == row["VARIEDAD"]) & (INFO_SUPERFICIE.MODO == row["MODO"]), "SUPERFICIE"])
            if len(sup_value) > 0:
                part1+=1
                if len(sup_value) > 1:
                    UH_2023_TRAIN.loc[i, "SUPERFICIE"] = np.mean(sup_value)
                else:
                    UH_2023_TRAIN.loc[i, "SUPERFICIE"] = sup_value[0]
            else:
                sup_value = list(INFO_SUPERFICIE.loc[(INFO_SUPERFICIE.ID_FINCA == row["ID_FINCA"]) & (INFO_SUPERFICIE.VARIEDAD == row["VARIEDAD"]), "SUPERFICIE"])
                if len(sup_value) > 0:
                    part2+=1
                    if len(sup_value) > 1:
                        UH_2023_TRAIN.loc[i, "SUPERFICIE"] = np.mean(sup_value)
                    else:
                        UH_2023_TRAIN.loc[i, "SUPERFICIE"] = sup_value[0]
                else:
                    sup_value = list(INFO_SUPERFICIE.loc[(INFO_SUPERFICIE.ID_FINCA == row["ID_FINCA"]), "SUPERFICIE"])
                    if len(sup_value) > 0:
                        part3+=1
                        if len(sup_value) > 1:
                            UH_2023_TRAIN.loc[i, "SUPERFICIE"] = np.mean(sup_value)
                        else:
                            UH_2023_TRAIN.loc[i, "SUPERFICIE"] = sup_value[0]
                    else:
                        UH_2023_TRAIN.loc[i, "SUPERFICIE"] = np.nan #Si no tenemos datos para la superficie de la finca, imputamos un NaN.
    else:
        pass
    

# Imputando los valores de la altitud, en base al ID de la estación meteorológica de la finca:
    
altitud = UH_2023_TRAIN.groupby("ID_ESTACION").agg({"ALTITUD":np.nanmedian}).reset_index()
for i, row in UH_2023_TRAIN.iterrows():
    if math.isnan(row["ALTITUD"]):
        alt_val = altitud.loc[altitud.ID_ESTACION == row["ID_ESTACION"], "ALTITUD"].values
        if len(alt_val) > 0:
            UH_2023_TRAIN.loc[i, "ALTITUD"] = alt_val[0]
        else:
            pass
    else:
        pass
    
#Creamos las variables de los retardos de la producción, para 1 y 2 años menos, respectivamente
lag_prod_values = []
for i, row in UH_2023_TRAIN.iterrows():
    campaña = row["CAMPAÑA"] - 1
    finca = row["ID_FINCA"]
    variedad = row["VARIEDAD"]
    lag_prod = UH_2023_TRAIN.loc[(UH_2023_TRAIN.CAMPAÑA == campaña) & (UH_2023_TRAIN.ID_FINCA == finca) & (UH_2023_TRAIN.VARIEDAD == variedad)
                                 , "PRODUCCION"].values
    if len(lag_prod) > 0:
        if len(lag_prod) == 1:
            lag_prod_values.append(lag_prod[0])
        else:
            lag_prod_values.append(np.mean(lag_prod))
    else:
        lag_prod_values.append(np.nan)
        
lag_prod_values_2 = []
for i, row in UH_2023_TRAIN.iterrows():
    campaña = row["CAMPAÑA"] - 2
    finca = row["ID_FINCA"]
    variedad = row["VARIEDAD"]
    lag_prod = UH_2023_TRAIN.loc[(UH_2023_TRAIN.CAMPAÑA == campaña) & (UH_2023_TRAIN.ID_FINCA == finca) & (UH_2023_TRAIN.VARIEDAD == variedad)
                                 , "PRODUCCION"].values
    if len(lag_prod) > 0:
        if len(lag_prod) == 1:
            lag_prod_values_2.append(lag_prod[0])
        else:
            lag_prod_values_2.append(np.mean(lag_prod))
    else:
        lag_prod_values_2.append(np.nan)
        
UH_2023_TRAIN["LAG_PROD"] = lag_prod_values
UH_2023_TRAIN["LAG2_PROD"] = lag_prod_values_2

#Continuamos con la transformación de los datos:

#Nos quedamos con el año, mes y día de la fecha
DATOS_METEO["Year"] = list(map(str,DATOS_METEO.validTimeUtc.dt.year))
DATOS_METEO["Month"] = list(map(str,DATOS_METEO.validTimeUtc.dt.month))
DATOS_METEO["Day"] = list(map(str,DATOS_METEO.validTimeUtc.dt.day))

#Mantenemos únicamente las columnas que consideramos importantes.
columns = ["precip1Hour", "Year", "Month", "Day", "pressureChange", "pressureMeanSeaLevel", 
          "relativeHumidity", "snow1Hour", "temperature", "temperatureDewPoint","temperatureFeelsLike",
          "uvIndex", "visibility", "windSpeed", "windDirection", "ID_ESTACION"] 

DATOS_METEO_FINAL = DATOS_METEO[[i for i in DATOS_METEO.columns if i in columns]].copy()

#Reemplazamos los valores de la columna uvIndex que marcan un NaN
DATOS_METEO_FINAL.uvIndex.replace([-2,-1], np.nan, inplace = True)


#Agregamos las filas de forma que obtengamos información diaria a partir de la información horaria originalmente disponible. Si nos interesa dar cierto peso a datos atípicos, usamos la media. Si no, la mediana. Finalmente, para cierto tipo de variables, mantenemos también la suma, el mínimo o el máximo, además de la moda para las variables que sean categóricas.
DATOS_METEO_FINAL = DATOS_METEO_FINAL.groupby(["ID_ESTACION", "Day", "Month", "Year"], sort = False).agg({"precip1Hour":[np.nanmean,np.nanmin,np.nanmax],
                                                                                      "relativeHumidity": np.nanmedian,
                                                                                      "snow1Hour": [np.nanmean,np.nanmin,np.nanmax], 
                                                                                      "temperature": [np.nanmedian, np.nanmax, np.nanmin],
                                                                                      })
DATOS_METEO_FINAL.columns = ["_".join(col) for col in DATOS_METEO_FINAL.columns.values]
DATOS_METEO_FINAL.columns = [i.split("_")[1][3:] + i.split("_")[0].capitalize().replace("1hour","") for i in DATOS_METEO_FINAL.columns.values] 
DATOS_METEO_FINAL = DATOS_METEO_FINAL.reset_index()

#Obtenemos el año, mes y día de los datos. Posteriormente, los transformamos para que tengan el formato adecuado.
DATOS_ETO["date"] = [i[:4] + "-" + i[4:6] + "-" + i[6:] for i in map(str,DATOS_ETO.date)]
DATOS_ETO[["Year", "Month", "Day"]] = DATOS_ETO.date.str.split("-", expand = True)
DATOS_ETO.loc[DATOS_ETO.Month.isin([i for i in DATOS_ETO.Month.unique() if int(i) <= 9]), "Month"] = [i.replace("0", "") for i in DATOS_ETO.Month if int(i) < 10]
DATOS_ETO.loc[DATOS_ETO.Day.isin([i for i in DATOS_ETO.Day.unique() if int(i) <= 9]), "Day"] = [i.replace("0", "") for i in DATOS_ETO.Day if int(i) < 10]
DATOS_ETO.drop("date", axis = 1, inplace = True)

# Juntamos los todos los datos meteorológicos en un solo dataframe:
DATOS_CLIMA = pd.merge(DATOS_METEO_FINAL, DATOS_ETO, how = "inner", on = ["ID_ESTACION", "Day", "Month", "Year"], suffixes = ("_METEO", "_ETO"))

#Mantenemos los datos de año, mes y día en el formato adecuado.
DATOS_CLIMA["Year"] = list(map(int, DATOS_CLIMA["Year"]))
DATOS_CLIMA["Month"] = list(map(int, DATOS_CLIMA["Month"]))
DATOS_CLIMA["Day"] = list(map(int, DATOS_CLIMA["Day"]))

#Obtenemos la fase del ciclo de la uva a la que pertenece cada observación, haciendo uso de la función definida al principio del script.
DATOS_CLIMA[["Phase", "PredYear"]] = pd.DataFrame(list(map(getPhase, [row for i,row in DATOS_CLIMA.iterrows()])))

#Eliminamos los datos relativos al 2015, ya que solamente tenemos dos datos, a saber, 29 y 30 de junio.
DATOS_CLIMA.drop(DATOS_CLIMA.loc[DATOS_CLIMA.PredYear == 2015].index, axis = 0, inplace = True)

columns = list(DATOS_CLIMA.columns)
#Mantenemos solamente las variables climáticas más importantes, relativas a la temperatura, humedad, lluvia, nieve y viento 
selected_columns = columns[:11] + columns[62:65] + columns[74:77] + columns[272:275] + columns[287:]

aggColumns = {i:np.nanmean for i in selected_columns if i not in ["ID_ESTACION", "Day", "Month", "Year", "PredYear", "Phase"]}
DATOS_CLIMA = DATOS_CLIMA.groupby(["ID_ESTACION", "PredYear", "Phase"]).agg(aggColumns).reset_index()

#Colocamos los valores en columnas con su nombre correspondiente.
values = [(est, int(float(str(year)[2:]))) for est in DATOS_CLIMA["ID_ESTACION"].unique() for year in DATOS_CLIMA["PredYear"].unique()]
data = {"ID_ESTACION":[i[0] for i in values], "CAMPAÑA":[i[1] for i in values]}
for phase in DATOS_CLIMA.Phase.unique():
    for column in [i for i in DATOS_CLIMA.columns if i not in ["ID_ESTACION", "PredYear", "Phase"]]:
        name = phase + "_" + column
        value = DATOS_CLIMA.loc[DATOS_CLIMA.Phase == phase, column].values
        data[name] = value
        
#Obtenemos el dataset meterológico final, con toda la información meteorológica en el formato adecuado.
DATOS_CLIMA_FINAL = pd.DataFrame(data = data)

cols = ["L","B","F1","F2","M","R"] #Mantenemos solamente las fases del ciclo de la vid más relevantes
DATOS_CLIMA_FINAL = DATOS_CLIMA_FINAL[[i for i in DATOS_CLIMA_FINAL.columns if True in [True for j in cols if j in i]] + ["ID_ESTACION"]]

# Obtenemos el dataset final, con la información de las fincas.
WINE_DATA = pd.merge(UH_2023_TRAIN, DATOS_CLIMA_FINAL, how = "inner", on = ["ID_ESTACION", "CAMPAÑA"], suffixes = ("_TRAIN", "CLIMA")).sort_values(["ID_FINCA", "CAMPAÑA"])

#Separamos los datos de 2022 del resto.
WINE_DATA_PRED = WINE_DATA.loc[WINE_DATA.PRODUCCION.isna(), :].copy()
WINE_DATA_TRAIN = WINE_DATA.loc[WINE_DATA.PRODUCCION.isna() == False, :].copy()

#Finalmente, enviamos el ID de la finca al índice y eliminamos la columna con el año de la campaña.
WINE_DATA_PRED.set_index("ID_FINCA", inplace = True)
WINE_DATA_TRAIN.set_index("ID_FINCA", inplace = True)
WINE_DATA_PRED.drop("CAMPAÑA", axis = 1, inplace = True)
WINE_DATA_TRAIN.drop("CAMPAÑA", axis = 1, inplace = True)

# Parte de train
X = WINE_DATA_TRAIN[[i for i in WINE_DATA_TRAIN.columns if i not in ["PRODUCCION", "ID_FINCA", "ID_ZONA", "ID_ESTACION"]]]
X = X.drop(columns = X.columns[X.apply(lambda x: np.std(x) == 0, axis = 0)])  #Eliminamos todas las columnas con varianza 0
#Creamos interacciones de la superficie con la variedad de la uva y el modo de cultivo
X["SUP_VAR"] = X["SUPERFICIE"]*X["VARIEDAD"] 
X["SUP_MOD"] = X["SUPERFICIE"]*X["MODO"]
#Eliminamos aquellas columnas que no son relevantes para producir
X.drop(["L_maxPrecip", "M_medianRelativehumidity", "F2_FeelsLikeLocalDaytimeMin", "F2_FeelsLikeLocalOvernightMin", "C_medianRelativehumidity"], axis = 1, inplace = True)

#Creamos los array de los datos
X_2 = np.array(X)
y_2 = np.array(WINE_DATA_TRAIN[["PRODUCCION"]])

#Hacemos el split entre train y test
X_train_2, X_test_2, y_train_2, y_test_2 = train_test_split(X_2,y_2,random_state = 669280, shuffle = True, test_size = 0.3)


# Parte de predicción
X_pred_2 = WINE_DATA_PRED[[i for i in WINE_DATA_PRED.columns if i not in ["PRODUCCION", "ID_FINCA", "ID_ZONA", "ID_ESTACION"]]]

#Repetimos el proceso llevado a cabo en los datos de entrenamiento
X_pred_2["SUP_VAR"] = X_pred_2["SUPERFICIE"]*X_pred_2["VARIEDAD"]
X_pred_2["SUP_MOD"] = X_pred_2["SUPERFICIE"]*X_pred_2["MODO"]
X_pred_2.drop(["L_maxPrecip", "M_medianRelativehumidity", "F2_FeelsLikeLocalDaytimeMin", "F2_FeelsLikeLocalOvernightMin", "C_medianRelativehumidity"], axis = 1, inplace = True)

X_pred_2 = X_pred_2[[i for i in X_pred_2.columns if i in X.columns]]

X_pred_2 = np.array(X_pred_2)

print(X.shape)

### CREACION DE FUNCIONES AUXILIARES PARA LA PIPELINE

# Esta clase permite implementar funciones creadas por el usuario en una Pipeline de sklearn.
class DataframeFunctionTransformer():
    def __init__(self, func):
        self.func = func

    def transform(self, input_df, **transform_params):
        return self.func(input_df)

    def fit(self, X, y=None, **fit_params):
        return self

def efficiency_diff(dataset):
    
    """
    Esta función recibe un array de numpy con las columnas adecuadas, y devuelve el mismo array con dos columnas adicionales, una con el logaritmo de la eficiencia de la finca, que se calcula como la cantidad producida entre la superficie y otra con las primeras diferencias para la producción del año anterior al de la fila.
    """
    
    #Obtenemos las columnas necesarias para crear las variables
    superficie = dataset[:,5]
    lag_prod = dataset[:,6]
    lag2_prod = dataset[:,7]
    
    #Creamos las variables y las ponemos en la dimensión adecuada para usar np.concatenate
    diff = lag_prod - lag2_prod
    diff = diff.reshape(diff.shape[0], 1)
    eficiencia = lag_prod/superficie
    eficiencia = np.log(eficiencia)
    eficiencia = eficiencia.reshape(eficiencia.shape[0], 1)
    
    #Concatenamos las nuevas columnas
    dataset = np.concatenate((dataset, diff), axis = 1)
    dataset = np.concatenate((dataset, eficiencia), axis = 1)
    return dataset

### AJUSTE DEL MODELO
"""
cv = RepeatedKFold(n_splits=10, n_repeats=3, random_state=seed_num) #Objeto cv para el k-fold cross-validation implementado posteriormente

#Definimos la función de búsqueda para Sequential Model Optimization (SMO):
def objective(trial):
    params = {
        'learning_rate' : trial.suggest_float("learning_rate",0.15, 0.3),
        'max_iter' : trial.suggest_int("max_iter",50,1500),
        'max_leaf_nodes' : trial.suggest_int("max_leaf_nodes",10,200),
        'max_depth' : trial.suggest_int("max_depth",5,15),
        'random_state': seed_num,
        'loss' : "squared_error"
    }

    model = Pipeline([
        ("ChildColumn", DataframeFunctionTransformer(efficiency_diff)), 
        ("reg",HistGradientBoostingRegressor(**params))])

    mean_rmse_score =  -np.mean(cross_val_score(model,X_2,y_2.ravel(),cv = cv, scoring = "neg_root_mean_squared_error"
                                        , error_score = "raise", n_jobs = 6))

    return mean_rmse_score

budget = 500
study_HGB_2 = optuna.create_study(study_name = "HGBHPO", direction = "minimize")
study_HGB_2.optimize(objective, n_trials = budget)

print("The best set of parameters is: ", study_HGB_2.best_params)
print("The rmse score computed with inner-evaluation is:", study_HGB_2.best_value)
print(f"The optimal value is reached in {study_HGB_2.best_trial.number} iterations")
"""
print(X.columns)

# Definimos el modelo
best_params_HGB = {'learning_rate': 0.16966582726087054, 'max_iter': 50, 'max_leaf_nodes': 14, 'max_depth': 6}
model_HGB_2 = Pipeline([
        ("imputer", IterativeImputer()),
        ("ChildColumn", DataframeFunctionTransformer(efficiency_diff)), # Obtener las columnas que queremos
        ("reg",HistGradientBoostingRegressor(**best_params_HGB))])

# Entrenamos
model_HGB_2.fit(X_2,y_2.ravel())

# Predecimos
y_pred_2 = model_HGB_2.predict(X_pred_2)


exp2 = dx.Explainer(model_HGB_2, X_2, np.squeeze(y_2))
mp2 = exp2.model_parts(loss_function = 'rmse')

# we can plot the results
mp2.plot()

relevant_var_idx2 = list(mp2.result.variable[-11:-1])
relevant_cols2 = [X.columns[int(idx)] for idx in relevant_var_idx2]
relevant_cols2.reverse()

print("The most relevant variables for the analysis of HGB are: ")
print(relevant_cols2)

sample_data = X_2[100,:]
rf_pparts = exp2.predict_parts(new_observation = sample_data, type = "break_down")
rf_pparts.plot()


protected_21 = np.where(X_2[:,2] == 1, "modo_1", "else")
protected_22 = np.where(X_2[:,2] == 2, "modo_2", "else")
protected_30 = np.where(X_2[:,3] == 0, "tipo_0", "else")
protected_31 = np.where(X_2[:,3] == 1, "tipo_1", "else")
protected_40 = np.where(X_2[:,4] == 0, "color_0", "else")
protected_41 = np.where(X_2[:,4] == 1, "color_1", "else")
privileged = "else"

fobject2 = exp2.model_fairness(protected_21, privileged)
fobject2.fairness_check()
fobject2 = exp2.model_fairness(protected_22, privileged)
fobject2.fairness_check()
fobject2 = exp2.model_fairness(protected_30, privileged)
fobject2.fairness_check()
fobject2 = exp2.model_fairness(protected_31, privileged)
fobject2.fairness_check()
fobject2 = exp2.model_fairness(protected_40, privileged)
fobject2.fairness_check()
fobject2 = exp2.model_fairness(protected_41, privileged)
fobject2.fairness_check()

# Ajustamos el formato
y_pred_2 = y_pred_2.flatten()
pred_data = WINE_DATA_PRED.reset_index()[["ID_FINCA","VARIEDAD","MODO","TIPO","COLOR","SUPERFICIE"]].copy()
pred_data["PRODUCCION"] = list(map(lambda x: round(x, 2), y_pred_2))
pred_data = pred_data.sort_values(["ID_FINCA","VARIEDAD","MODO","TIPO","COLOR","SUPERFICIE","PRODUCCION"])

#Obtenemos el .txt final.
pred_data.to_csv("E-No-Logos_500.txt", header = False, sep='|', index = False)